//#define LOG_NDEBUG 0
#define LOG_TAG "CameraHal3"

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/sysmacros.h>
#include <sys/stat.h>
#include <log/log.h>
#include <utils/threads.h>
#include <utils/Vector.h>
#include <utils/CallStack.h>
#include <linux/videodev2.h>
#include <binder/MemoryBase.h>
#include <binder/MemoryHeapBase.h>
#include <CameraParameters.h>
#include <hardware/camera3.h>
#include <cutils/properties.h>
#include "Debug.h"
#include "CameraFactory.h"
#ifdef CAMERA_MONITOR
#include "CameraMonitor.h"
#endif

#define DEVICE_API_VERSION  CAMERA_DEVICE_API_VERSION_3_3

// camera facing properties //XCheck: needed or not.
#define USBCAM_FACING_PROPERTY      "persist.vendor.rtk.camera.facing"

#define FACING_DEFAULT              "default"
#define FACING_FRONT                "front"
#define FACING_BACK                 "back"
#define FACING_EXTERNAL             "external"
#define FACING_NONE                 "none"

// #define USBCAM_DEFAULT_FACING       CAMERA_FACING_EXTERNAL // XCheck
#define USBCAM_DEFAULT_FACING       CAMERA_FACING_BACK

#define CAMERA_NUM  1
#define USB_ID      0

using namespace android;
#ifdef CAMERA_MONITOR
using namespace android_camera_hal3;
#endif
using namespace androidcamera3;

typedef enum {
    STATE_NO_CHANGE,
    PRESENT_TO_NOT_PRESENT,
    NOT_PRESENT_TO_PRESENT
} camera_state_change_enum;

static const camera_module_callbacks_t* gCallbacks;
static int camera_device_close(hw_device_t* device);

typedef struct CAMERA_INFO_RTK {

    int cameraId;
    int v4l2devId;
    int facing;
    int state;

    CameraHardware3Base* pHw;

    CAMERA_INFO_RTK()
        : cameraId(-1),
          v4l2devId(-1),
          facing(-1),
          state(CAMERA_DEVICE_STATUS_NOT_PRESENT),
          pHw(NULL)
    { }

} camera_info_rtk_t;

// TODO: review this later
typedef struct V4L2_camera3_device {
    struct camera3_device base;
    int cameraid;
    //CameraHardware3 *hw;
    camera_info_rtk_t* pPrivHandle;
} V4l2_camera3_device_t;

static camera_info_rtk_t gCameras[1];

typedef struct _FACING_INFO_STRUCT {
    int mFacing;
    bool mExistance;
    bool mInited;
} FACING_INFO_STRUCT;

static FACING_INFO_STRUCT sUsbCamFacingInfo = {USBCAM_DEFAULT_FACING, true, false};

static const char* get_cam_facing_str(int facing)
{
    switch(facing) {
        case CAMERA_FACING_FRONT:
            return "facing_front";
        case CAMERA_FACING_BACK:
            return "facing_back";
        case CAMERA_FACING_EXTERNAL:
            return "facing_external";
        default:
            return "facing_undefine";
    }
}

static void resolve_facing_setting()
{
    if (sUsbCamFacingInfo.mInited) {
        return;
    }

    if (!sUsbCamFacingInfo.mInited) {
        sUsbCamFacingInfo.mInited = true;

        char value[PROPERTY_VALUE_MAX];
        memset(value, 0x0, PROPERTY_VALUE_MAX);
        property_get(USBCAM_FACING_PROPERTY, value, FACING_DEFAULT);

        sUsbCamFacingInfo.mExistance = true;

        if (0 == strcmp(value, FACING_FRONT)) {
            sUsbCamFacingInfo.mFacing = CAMERA_FACING_FRONT;
        } else if (0 == strcmp(value, FACING_BACK)) {
            sUsbCamFacingInfo.mFacing = CAMERA_FACING_BACK;
        } else if (0 == strcmp(value, FACING_EXTERNAL)) {
            sUsbCamFacingInfo.mFacing = CAMERA_FACING_EXTERNAL;
        } else if (0 == strcmp(value, FACING_DEFAULT)) {
            sUsbCamFacingInfo.mFacing = USBCAM_DEFAULT_FACING;
        } else if (0 == strcmp(value, FACING_NONE)) {
            sUsbCamFacingInfo.mFacing = USBCAM_DEFAULT_FACING;
            sUsbCamFacingInfo.mExistance = false;
        }

        ALOGV("USBCam facingInfo %i(%s) existance %d", sUsbCamFacingInfo.mFacing, get_cam_facing_str(sUsbCamFacingInfo.mFacing), sUsbCamFacingInfo.mExistance);
    }
}

static inline bool isCameraAvail(int minor)
{
    struct stat st;
    char devnode[32];
    sprintf(devnode, "/dev/video%d", minor);
    if (stat(devnode, &st) != 0) {
        return false;
    }

    ALOGV("found device %s chr:%d major:%d", devnode, S_ISCHR(st.st_mode), major(st.st_rdev));
    if (!S_ISCHR(st.st_mode) || major(st.st_rdev) != VIDEO_MAJOR) {
        return false;
    }

    ALOGV("found V4L2 device: %s", devnode);
    return true;
}

static inline int findUSBCamDevId(void)
{
    //video0 ~ video63 are video capture/overlay devices
    for (int i = 0; i < 4; i++) {
        if (isCameraAvail(i)) {
            ALOGV("[CAMUPDATE] Found a USB camera device %d!", i);
            return i;
        }
    }
    ALOGV("[CAMUPDATE] No USB camera device found!!");
    return -1;
}

static inline void camera_check_camera(int id, camera_state_change_enum* report)
{
    LOG_FUNCTION_NAME();
    CHECK((id >= 0 && id <= 1));
    camera_info_rtk_t* pCam = &(gCameras[id]);

    *report = STATE_NO_CHANGE;

    ALOGV("%s[%d]: [CAMUPDATE] id %d current state: %d", __FUNCTION__, __LINE__, id, pCam->state);
    if(pCam->state == CAMERA_DEVICE_STATUS_PRESENT) {
        if(!isCameraAvail(pCam->v4l2devId)) {
            // remove camera
            pCam->cameraId = USB_ID;
            pCam->v4l2devId = -1;
            pCam->facing = sUsbCamFacingInfo.mFacing;
            pCam->state = CAMERA_DEVICE_STATUS_NOT_PRESENT;
            if(pCam->pHw != NULL) {
                pCam->pHw->cleanupOnPlugout();
            }
            *report = PRESENT_TO_NOT_PRESENT;
            ALOGV("%s[%d]: [CAMUPDATE] USB_ID set report to PRESENT_TO_NOT_PRESENT", __FUNCTION__, __LINE__);
        }
    } else { // CAMERA_DEVICE_STATUS_NOT_PRESENT
        int usbV4L2devId = findUSBCamDevId();
        if(usbV4L2devId >= 0 && sUsbCamFacingInfo.mExistance) {
            pCam->cameraId = USB_ID;
            pCam->v4l2devId = usbV4L2devId;
            pCam->facing = sUsbCamFacingInfo.mFacing;
            pCam->state = CAMERA_DEVICE_STATUS_PRESENT;
            if(pCam->pHw != NULL) {
                delete pCam->pHw;
                pCam->pHw = NULL;
            }

            pCam->pHw = androidcamera3::createCamera(USB_ID, pCam->v4l2devId, pCam->facing); // XCheck: USB_ID parameter
            // pCam->pHw = dynamic_cast<CameraHardware3Base *>(new CameraHardware3v2(USB_ID, pCam->v4l2devId, pCam->facing));
            *report = NOT_PRESENT_TO_PRESENT;
            ALOGV("%s[%d]: [CAMUPDATE] USB_ID set report to NOT_PRESENT_TO_PRESENT", __FUNCTION__, __LINE__);
        }
    }
}

static inline int camera_init_cam_info_t(const camera_module_callbacks_t* callbacks)
{
    LOG_FUNCTION_NAME();
    camera_state_change_enum report;
    int rst = 0;

    camera_check_camera(USB_ID, &report);

    ALOGV("%s[%d]: [CAMUPDATE] report %d", __FUNCTION__, __LINE__, report);

    // notify camera framework that camera is available or not
    if(callbacks != NULL) {
        if(report == PRESENT_TO_NOT_PRESENT) {
            ALOGV("%s[%d]: [CAMUPDATE] camera 0 CAMERA_DEVICE_STATUS_NOT_PRESENT", __FUNCTION__, __LINE__);
            callbacks->camera_device_status_change(callbacks, USB_ID, CAMERA_DEVICE_STATUS_NOT_PRESENT);
            rst = 1;
        } else if(report == NOT_PRESENT_TO_PRESENT) {
            ALOGV("%s[%d]: [CAMUPDATE] camera 0 CAMERA_DEVICE_STATUS_PRESENT", __FUNCTION__, __LINE__);
            callbacks->camera_device_status_change(callbacks, USB_ID, CAMERA_DEVICE_STATUS_PRESENT);
            rst = 1;
        }
    }

    return rst;
}

static inline int getNumberOfCameras()
{
#ifdef CAMERA_MONITOR
#else
    for (int i = 0; i < 20; i++) { // 20 times
        int usbV4L2devId = findUSBCamDevId();
        if (usbV4L2devId >= 0) {
            ALOGV("%s[%d]: return number: %d", __FUNCTION__, __LINE__, CAMERA_NUM);
            return CAMERA_NUM;  // Currently we only support 1 usb camera.
        }
        usleep(500000); // 500ms
    }
    ALOGD("%s[%d]: return number: 0", __FUNCTION__, __LINE__);
#endif
    return 0;
}

//CameraService limits max # of cameras to 2
int camera_get_number_of_cameras(void)
{
    LOG_FUNCTION_NAME();
    resolve_facing_setting();
    return getNumberOfCameras();
}

static int camera_refresh_cameras()
{
    LOG_FUNCTION_NAME();
    for(int i = 0; i < 10; i++) {
        if(camera_init_cam_info_t(gCallbacks) > 0) {
            ALOGV("[CAMUPDATE] camera_refresh_cameras finished at round [%d]", i);
            break;
        }
        usleep(5000);
    }
    return 0;
}

int camera_set_callbacks(const camera_module_callbacks_t* callbacks)
{
    LOG_FUNCTION_NAME();
    gCallbacks = callbacks;
#ifdef CAMERA_MONITOR
    ALOGV("%s[%d]: [CAMUPDATE] camera 0 CAMERA_DEVICE_STATUS_NOT_PRESENT", __FUNCTION__, __LINE__);
    callbacks->camera_device_status_change(callbacks, USB_ID, CAMERA_DEVICE_STATUS_NOT_PRESENT);
#endif
    camera_init_cam_info_t(gCallbacks);

#ifdef CAMERA_MONITOR
    camera_state_change_enum report;
    camera_check_camera(USB_ID, &report);
    CameraMonitor::Instance()->setCallback(&camera_refresh_cameras);
#endif

    return 0;
}

/**
 * metadata tag info can refer to
 * /system/media/camera/include/system/camera_metadata_tags.h
 * and
 * /system/media/camera/src/camera_metadata_tag_info.c
 *   from camera_common.h
 *   When a camera is disconnected, its camera id becomes invalid. Calling this
 *   this method with this invalid camera id will get -EINVAL and NULL camera
 *   static metadata (camera_info.static_camera_characteristics).
 */
int camera_get_camera_info(int cameraId, struct camera_info* info)
{
    LOG_FUNCTION_NAME();
    info->orientation = 0;
    info->facing = gCameras[cameraId].facing;
    info->device_version = DEVICE_API_VERSION; // Hal3 fields

    if(gCameras[cameraId].pHw != NULL) {
        info->static_camera_characteristics = gCameras[cameraId].pHw->getStaticInfo();
    } else {
        info->static_camera_characteristics = NULL;
    }

    info->resource_cost = 10;
    info->conflicting_devices = NULL;
    info->conflicting_devices_length = 0;
    ALOGV("%s[%d]: cameraId:%d orientation:%d facing:%d", __FUNCTION__, __LINE__, cameraId, info->orientation, info->facing);

    // for pluggable device work-around
    if (gCameras[cameraId].state == CAMERA_DEVICE_STATUS_NOT_PRESENT) {
        ALOGV("%s[%d]: camera %d still not ready", __FUNCTION__, __LINE__, cameraId);
        return -EINVAL; //-ENODEV;
    } else {
        ALOGV("%s[%d]: camera %d ready", __FUNCTION__, __LINE__, cameraId);
    }

    return 0;
}

int camera_device_close(hw_device_t* device)
{
    LOG_FUNCTION_NAME();
    V4l2_camera3_device_t* dev = (V4l2_camera3_device_t*)device;
    if(dev->pPrivHandle->pHw != NULL) {

        ALOGV("%s[%d]: closeCamera..", __FUNCTION__, __LINE__);

        dev->pPrivHandle->pHw->closeCamera();
        dev->pPrivHandle->pHw->uninitialize();
    }
    return 0;
}

// ================================ camera3 Hal callback ===============================
static int initialize(const struct camera3_device* _dev, const camera3_callback_ops_t* callback_ops)
{
    LOG_FUNCTION_NAME();
    V4l2_camera3_device_t* dev = (V4l2_camera3_device_t*)_dev;
    if(dev->pPrivHandle->pHw != NULL) {
        return dev->pPrivHandle->pHw->initialize(callback_ops);
    }
    return 0;
}

static int configure_streams(const struct camera3_device* _dev, camera3_stream_configuration_t* stream_list)
{
    LOG_FUNCTION_NAME();
    V4l2_camera3_device_t* dev = (V4l2_camera3_device_t*)_dev;
    if(dev->pPrivHandle->pHw != NULL) {
        return dev->pPrivHandle->pHw->configureStreams(stream_list);
    }

    return 0;
}

static const camera_metadata_t* construct_default_request_settings(const struct camera3_device* _dev, int type)
{
    LOG_FUNCTION_NAME();
    V4l2_camera3_device_t* dev = (V4l2_camera3_device_t*)_dev;
    camera_metadata_t* rst = NULL;
    if(dev->pPrivHandle->pHw != NULL) {
        rst = dev->pPrivHandle->pHw->translateCapabilityToMetadata(type);
    }

    return rst;
}

static int process_capture_request(const struct camera3_device* _dev, camera3_capture_request_t* request)
{
    LOG_FUNCTION_NAME();
    V4l2_camera3_device_t* dev = (V4l2_camera3_device_t*)_dev;
    if(dev->pPrivHandle->pHw != NULL) {
        return dev->pPrivHandle->pHw->processCaptureRequest(request);
    }

    return 0;
}

static void dump(const struct camera3_device*, int fd)
{
    LOG_FUNCTION_NAME();
    const char* str = "Realtek Camera Hal3";
    write(fd, str, strlen(str));
    fsync(fd);
}

static int flush(const struct camera3_device* _dev)
{
    LOG_FUNCTION_NAME();
    V4l2_camera3_device_t* dev = (V4l2_camera3_device_t*)_dev;
    if(dev->pPrivHandle->pHw != NULL) {
        return dev->pPrivHandle->pHw->flush();
    }

    return 0;
}
// ================================ camera3 Hal callback ===============================

int camera_device_open(const hw_module_t* module, const char* name, hw_device_t** hwdev)
{
    LOG_FUNCTION_NAME();
    int i;
    int rv = 0;
    int cameraid;
    V4l2_camera3_device_t* dev = NULL;
    camera3_device_ops_t* camera_ops = NULL;

    if (name != NULL) {
        cameraid = atoi(name);
        if (cameraid < 0 || cameraid >= 2) {
            ALOGE("Invalid camera id:%d", cameraid);
            rv = -EINVAL;
            goto fail;
        }

        camera_info_rtk_t* pCam = &(gCameras[cameraid]);

        if (pCam->v4l2devId == -1 || isCameraAvail(pCam->v4l2devId) == false) {
            ALOGE("Invalid camera device for id %d", cameraid);
            rv = -EINVAL;
            goto fail;
        }

        dev = (V4l2_camera3_device_t*)malloc(sizeof(*dev));
        if (NULL == dev) {
            ALOGE("V4l2_camera3_device_t allocation fail");
            rv = -ENOMEM;
            goto fail;
        }

        camera_ops = (camera3_device_ops_t*)malloc(sizeof(*camera_ops));
        if (NULL == camera_ops) {
            ALOGE("camera_ops allocation fail");
            rv = -ENOMEM;
            goto fail;
        }

        memset(dev, 0, sizeof(*dev));
        memset(camera_ops, 0, sizeof(*camera_ops));

        // setup dev
        dev->base.common.tag = HARDWARE_DEVICE_TAG;
        dev->base.common.version = DEVICE_API_VERSION;
        dev->base.common.module = (hw_module_t*)(module);
        dev->base.common.close = camera_device_close;
        dev->base.ops = camera_ops;
        // TODO: try to use priv field.

        // setup camera_ops
        camera_ops->initialize = initialize;
        camera_ops->configure_streams = configure_streams;
        camera_ops->register_stream_buffers = NULL;
        camera_ops->construct_default_request_settings = construct_default_request_settings;
        camera_ops->process_capture_request = process_capture_request;
        camera_ops->get_metadata_vendor_tag_ops = NULL;
        camera_ops->dump = dump;
        camera_ops->flush = flush;

        *hwdev = &dev->base.common;

        CHECK(pCam->pHw != NULL);

        dev->cameraid = cameraid;
        dev->pPrivHandle = &(gCameras[cameraid]);
        dev->pPrivHandle->pHw->connectCamera();
    }

    return rv;

fail:
    if (dev) {
        free(dev);
        dev = NULL;
    }

    if (camera_ops) {
        free(camera_ops);
        camera_ops = NULL;
    }
    return rv;
}
