//#define LOG_NDEBUG 0
#define LOG_TAG "CameraHardware3V2"

#include <cutils/properties.h>
#include <ui/Fence.h>
#include <sw_sync.h>
#include "CameraHardware3v2.h"
#include "V4L2Device2.h"
#include "V4L2Device2MJPG.h"
#include "Debug.h"

/**
 * one can enable/disable following options for debug purpose
 */
#define ADJUST_RESULT_FLOW      (1)

namespace android
{

static const nsecs_t kWaitNs = (5 * 1E9);

CameraHardware3v2::CameraHardware3v2(
    int camId,
    int v4l2Id,
    int facing)
    : CameraHardware3Base(camId, v4l2Id, facing),
      mDevice(NULL),
      mFlushed(false),
      mV4L2Params(),
      //mErrorFrames(0),
      mResultFrameNumber(0),
      mPrevSystemTime(0),
      mPrevPTS(0),
      mShutterThread(NULL),
      mFormatConvertHelper(NULL)
{
    LOG_FUNCTION_NAME();
    memset(&mV4L2Params, 0x0, sizeof(V4L2Params));
    mFrameNumberQueue.setCapacity(64);
    mInputRequestQueue.setCapacity(64);
    mOutputRequestQueue.setCapacity(64);
#ifdef HAL3_HWC_BUFFER
    mFormatConvertHelper = new FormatConvertHelper(pGrallocWrapper);
#else
    mFormatConvertHelper = new FormatConvertHelper();
#endif
}

CameraHardware3v2::~CameraHardware3v2()
{
    LOG_FUNCTION_NAME();
    closeCamera();

    if(mFormatConvertHelper != NULL) {
        delete mFormatConvertHelper;
        mFormatConvertHelper = NULL;
    }
}

status_t CameraHardware3v2::connectCamera(/*hw_device_t** device*/)
{
    LOG_FUNCTION_NAME();
    mFlushed = false;
    mResultFrameNumber = 0;

    Mutex::Autolock l(mLock);
    mShutterThread = new ShutterThread(this);
    mShutterThread->run("ShutterThread");

    status_t rst = connectCameraBase();
    if(rst != NO_ERROR) {
        return rst;
    }

    return NO_ERROR;
}

status_t CameraHardware3v2::closeCamera()
{
    LOG_FUNCTION_NAME();
    Mutex::Autolock l(mLock);

    if(mDevice != NULL) {
        mDevice->dumpDeviceStatus();
        delete mDevice;
        mDevice = NULL;
    }
    removeShutterThread();
    {
        Mutex::Autolock l(mFrameNumberQueueLock);
        int qs = mFrameNumberQueue.size();
        if(qs > 0) {
            ALOGW("%s[%d]: mFrameNumberQueue %d", __FUNCTION__, __LINE__, qs);
            for (int i = 0; i < qs; i++) {
                ALOGV("%s[%d]: \tPending Frames %d", __FUNCTION__, __LINE__, mFrameNumberQueue[i]);
            }
        }
    }

    closeCameraBase();

    mFlushed = false;
    mDoFlush = false;

    return NO_ERROR;
}

status_t CameraHardware3v2::configureStreams(camera3_stream_configuration_t* streamList)
{
    LOG_FUNCTION_NAME();
    Mutex::Autolock l(mLock);
    status_t res = configureStreamsBase(streamList);
    if (res != NO_ERROR) {
        return res;
    }

    ALOGV("%s[%d]: stream [%dx%d]", __FUNCTION__, __LINE__, stream_width, stream_height);
    // for Android O
    if(stream_width <= 0 || stream_height <= 0) {
        ALOGE("%s[%d]: bad stream dimension", __FUNCTION__, __LINE__);
        return -EINVAL;
    }

    CHECK(0 != stream_width);
    CHECK(0 != stream_height);

#if 0

    mV4L2Params.w = stream_width;
    mV4L2Params.h = stream_height;
    mV4L2Params.fps = mPerferedFps;
    mV4L2Params.v4l2format = mFormat;

    setV4L2Device(&mV4L2Params);
#endif
    return OK;
}

status_t CameraHardware3v2::processCaptureRequest(const Request& r)
{
    LOG_FUNCTION_NAME();
    Mutex::Autolock l(mLock);
    mRequestFrameNumber = r.frameNumber;
    collectRequestMiscInfo((CameraMetadata&) r.settings, (MISC_FIELD_INFO*) & (r.mMiscInfo));

    ALOGV("%s[%d]: frameNumber:%d:%s", __FUNCTION__, __LINE__, r.frameNumber, ((r.input_buffer == NULL) ? "Output" : "Input"));
    queueFrameNumber(r.frameNumber);

    if(r.input_buffer != NULL) {
        ALOGV("%s[%d]: input request %d", __FUNCTION__, __LINE__, r.frameNumber);
        queueReprocessRequest(r);
        return OK;
    }

    // TODO: if mDevice is deleted, also put to queue?
    if (!queueCaptureRequest(r)) {
        reportFailedRequest(r);
    }

    return OK;
}

bool CameraHardware3v2::queueCaptureRequest(const Request& r)
{
    LOG_FUNCTION_NAME();
    ALOGV("%s[%d]: frameNumber: %d", __FUNCTION__, __LINE__, r.frameNumber);
    if(NULL != mDevice) {
        ALOGV("%s[%d]: Call enqueueRequest %d", __FUNCTION__, __LINE__, r.frameNumber);
        mDevice->enqueueRequest(r);
        return true;
    } else {
        return false;
    }
}

void CameraHardware3v2::reportFailedRequest(const Request& r)
{
    LOG_FUNCTION_NAME();
    ALOGV("%s[%d]: frameNumber: %d", __FUNCTION__, __LINE__, r.frameNumber);

    Request* rr = new Request();
    *rr = r;
    rr->mSuccess = false;
    returnRequest(rr);
}

void CameraHardware3v2::removeShutterThread()
{
    // stop ShutterThread
    if(NULL != mShutterThread.get()) {
        mShutterThread->stop();
        mShutterThread->requestExitAndWait();
        mShutterThread.clear();
        mShutterThread = NULL;
    }
}

status_t CameraHardware3v2::flush()
{
    LOG_FUNCTION_NAME();
    mFlushed = true;

    Mutex::Autolock l(mLock);

    /**
     * we don't need flush to release enqueued buffers in V4L2 after update from kernel driver
     * unless device has been plugged out already.
     */
    if(mDoFlush) {

        if(mDevice != NULL) {
            mDevice->flush(true);
            delete mDevice;
            mDevice = NULL;
        }


        while(mOutputRequestQueue.size() > 0) {
            mShutterThread->signal();
            usleep(10);
        }

        CHECK(mOutputRequestQueue.size() == 0);

        ALOGV("%s[%d]: mStatus %d", __FUNCTION__, __LINE__, mStatus);
        mStatus = STATUS_READY;
    }
    return NO_ERROR;
}

status_t CameraHardware3v2::runtimeSetV4L2Device()
{
    LOG_FUNCTION_NAME();
    // choice fps
    int maxRuntimeFps = mRuntimeFps0;
    if(maxRuntimeFps < mRuntimeFps1) {
        maxRuntimeFps = mRuntimeFps1;
    }

    if(maxRuntimeFps == 0) {
        maxRuntimeFps = mPerferedFps;
    }

    mV4L2Params.w = stream_width;
    mV4L2Params.h = stream_height;
    mV4L2Params.fps = maxRuntimeFps;
    mV4L2Params.v4l2format = mFormat;

    ALOGV("%s[%d]: [%dx%d] fps:%d, fmt: %c%c%c%c", __FUNCTION__, __LINE__, mV4L2Params.w, mV4L2Params.h, mV4L2Params.fps, FCC2ASCII(mV4L2Params.v4l2format));
    setV4L2Device(&mV4L2Params);

    return OK;
}

void CameraHardware3v2::setV4L2Device(V4L2Params* params)
{
    LOG_FUNCTION_NAME();
    Mutex::Autolock l(mLock);

    if (NULL != mDevice) {
#if HAL3_HWC_BUFFER
        // then check if need to renew V4L2Device
        if ((0 != memcmp(params, mDevice->getParams(), sizeof(V4L2Params)))
            || !(mDevice->isStreaming())) {
            ALOGD("%s delete mDevice start", __FUNCTION__);
            mDevice->flush(true);
            delete mDevice;
            mDevice = NULL;
            ALOGD("%s delete mDevice done", __FUNCTION__);
        }
#else  // XChecked: Use runtime set parameters
        // then check if need to renew V4L2Device
        V4L2Params* devParams = mDevice->getParams();

        // XChecked: Here not check format before support MJPG format.
        bool bParamsUpdate = (params->w != devParams->w) || (params->h != devParams->h) || (params->fps != devParams->fps);
        bool bIsStreaming = mDevice->isStreaming();
        ALOGV("%s[%d]: [%dx%d] fps:%d, fmt:%c%c%c%c, bParamsUpdate:%i, bIsStreaming:%i", __FUNCTION__, __LINE__, devParams->w, devParams->h, devParams->fps, FCC2ASCII(devParams->v4l2format), bParamsUpdate, bIsStreaming);
        if (bParamsUpdate) {
            ALOGV("%s[%d]: delete mDevice start" , __FUNCTION__, __LINE__);
            mDevice->flush(true);
            delete mDevice;
            mDevice = NULL;
        }
#endif
    }

    if (NULL == mDevice) {
#if HAL3_HWC_BUFFER // XChecked: only support YUYV tmp. It will support MJPEG later.
        if(params->v4l2format != V4L2_PIX_FMT_MJPEG && params->v4l2format != V4L2_PIX_FMT_H264)
#endif
        {
            ALOGV("%s[%d]: v4l2format: %c%c%c%c, Create V4L2Device2 none MJPG format", __FUNCTION__, __LINE__, FCC2ASCII(params->v4l2format));
            mDevice = dynamic_cast<V4L2Device2Common*>(new V4L2Device2(
                          mV4l2Id,
                          params,
                          this,
                          kMaxBufferCount,
#ifdef HAL3_HWC_BUFFER
                          pGrallocWrapper,
#endif
                          mFormatConvertHelper,
                          false, //(mInputStream != NULL),
                          true));
#ifdef HAL3_HWC_BUFFER
        } else {
            ALOGV("%s[%d]: v4l2format: %c%c%c%c, Create V4L2Device2MJPG format", __FUNCTION__, __LINE__, FCC2ASCII(params->v4l2format));
            mDevice = dynamic_cast<V4L2Device2Common*>(new V4L2Device2MJPG(
                          mV4l2Id,
                          params,
                          this,
                          kMaxBufferCount,
                          pGrallocWrapper,
                          mFormatConvertHelper,
                          false,
                          false));
#endif
        }
        mDevice->init();
        mDevice->start();
    }
}

void CameraHardware3v2::submitErrorRequest(Request* r)
{
    LOG_FUNCTION_NAME();
    ALOGV("%s[%d]: frameNumber: %d", __FUNCTION__, __LINE__, r->frameNumber);
    mResultFrameNumber = r->frameNumber;

    HalBufferVector::iterator buf = r->buffers->begin();
    while(buf != r->buffers->end()) {
        buf->status = CAMERA3_BUFFER_STATUS_ERROR;
        // release fence
        // CTSTODO : review later
        buf->acquire_fence = -1;
        buf->release_fence = -1;

        ++buf;
    }

    nsecs_t captureTime = 0x2379;
    r->settings.update(ANDROID_SENSOR_TIMESTAMP,
                       &captureTime, 1);

    camera3_capture_result result;
    memset(&result, 0x0, sizeof(camera3_capture_result));

    result.frame_number = r->frameNumber;
    result.result = r->settings.getAndLock();
    result.num_output_buffers = r->buffers->size();
    result.output_buffers = r->buffers->array();
    result.input_buffer = r->input_buffer;
    result.partial_result = 1;

    mOps->process_capture_result(mOps, &result);

    camera3_notify_msg_t msg;
    msg.type = CAMERA3_MSG_SHUTTER;
    msg.message.shutter.frame_number = r->frameNumber;
    msg.message.shutter.timestamp = captureTime;

    mOps->notify(mOps, &msg);

    if(r->buffers != NULL) {
        delete r->buffers;
        r->buffers = NULL;
    }

    if(r->sensorBuffers != NULL ) {
        delete r->sensorBuffers;
        r->sensorBuffers = NULL;
    }

    // clean up
    r->settings.unlock(result.result);
    r->settings.clear();

    {
        Mutex::Autolock l(mFrameNumberQueueLock);
        int32_t firstFrameNumber = mFrameNumberQueue[0];
        CHECK(firstFrameNumber == (int32_t)r->frameNumber);
    }

    dequeueFrameNumber();
}

void CameraHardware3v2::submitRequest(Request* r)
{
    LOG_FUNCTION_NAME();
    ALOGV("%s[%d]:  frameNumber: %d, buffers %p", __FUNCTION__, __LINE__, r->frameNumber, r->buffers);
    mResultFrameNumber = r->frameNumber;

    bool needJpeg = (r->num_jpeg_output_buffers > 0) ? true : false;
    ALOGV("%s[%d]: needJpeg %d", __FUNCTION__, __LINE__, needJpeg);

    // release buffer Fence
    HalBufferVector::iterator buf = r->buffers->begin();
    int ii = 0;
    while (buf != r->buffers->end()) {
        bool bufValided = (*(r->sensorBuffers))[ii].bufferSuccess;
        int status = bufValided ? CAMERA3_BUFFER_STATUS_OK : CAMERA3_BUFFER_STATUS_ERROR;

        buf->status = status;
        // release fence
        buf->acquire_fence = -1;
        buf->release_fence = -1;
        ++buf;
        ii = ii + 1;
    } // end while

    ALOGV("%s[%d]: needJpeg:%d", __FUNCTION__, __LINE__, needJpeg);
    static const uint8_t sceneFlicker = ANDROID_STATISTICS_SCENE_FLICKER_NONE;
    r->settings.update(ANDROID_STATISTICS_SCENE_FLICKER, &sceneFlicker, 1);

    static const uint8_t flashState = ANDROID_FLASH_STATE_UNAVAILABLE;
    r->settings.update(ANDROID_FLASH_STATE, &flashState, 1);

    nsecs_t rollingShutterSkew = 33331760L; //Sensor::kFrameDurationRange[0];
    r->settings.update(ANDROID_SENSOR_ROLLING_SHUTTER_SKEW, &rollingShutterSkew, 1);

    float focusRange[] = { 1.0f / 5.0f, 0 }; // 5 m to infinity in focus
    r->settings.update(ANDROID_LENS_FOCUS_RANGE, focusRange, sizeof(focusRange) / sizeof(float));

    r->settings.update(ANDROID_SENSOR_TIMESTAMP, &(r->pts), 1);
    r->settings.update(ANDROID_COLOR_CORRECTION_TRANSFORM, colorTransform, 9);
    r->settings.update(ANDROID_COLOR_CORRECTION_GAINS, colorGains, 4);

    const uint8_t pipelineDepth = needJpeg ? kMaxBufferCount : kMaxBufferCount - 1;
    r->settings.update(ANDROID_REQUEST_PIPELINE_DEPTH, &pipelineDepth, 1);

    // send result
    camera3_capture_result result;
    memset(&result, 0x0, sizeof(camera3_capture_result));
    result.frame_number = r->frameNumber;
    result.result = r->settings.getAndLock();
    result.num_output_buffers = r->buffers->size();
    ALOGV("%s[%d]: num_output_buffers:%d", __FUNCTION__, __LINE__, result.num_output_buffers);
    result.output_buffers = r->buffers->array();
    result.input_buffer = nullptr;
    result.partial_result = 1;
    mOps->process_capture_result(mOps, &result);

    // notify shutter
    camera3_notify_msg_t msg;
    msg.type = CAMERA3_MSG_SHUTTER;
    msg.message.shutter.frame_number = r->frameNumber;
    msg.message.shutter.timestamp = r->pts;
    mOps->notify(mOps, &msg);

    // clear buffers
    if(r->buffers != NULL) {
        delete r->buffers;
        r->buffers = NULL;
    }

    // clear sensorBuffers
    if(r->sensorBuffers != NULL ) {    //if (!needJpeg) {
        delete r->sensorBuffers;
        r->sensorBuffers = NULL;
    }

    // resource clean up
    r->settings.unlock(result.result);
    r->settings.clear();

    int32_t firstFrameNumber = mFrameNumberQueue[0];
    CHECK(firstFrameNumber == (int32_t)r->frameNumber);
    dequeueFrameNumber();
}

void CameraHardware3v2::returnRequest(Request* r)
{
    LOG_FUNCTION_NAME();
    ALOGV("%s[%d]: \tframeNumber: %d", __FUNCTION__, __LINE__, r->frameNumber);
    queueFinishedOutputRequest(r);
}

void CameraHardware3v2::queueFinishedOutputRequest(Request* r)
{
    LOG_FUNCTION_NAME();
    {
        Mutex::Autolock l(mOutputRequestQueueLock);
        mOutputRequestQueue.push_back(r);
    }
    ALOGV("%s[%d]: \tframeNumber: %d", __FUNCTION__, __LINE__, r->frameNumber);
    mShutterThread->signal();
}

void CameraHardware3v2::dequeueOutputRequest()
{
    LOG_FUNCTION_NAME();
    Mutex::Autolock l(mOutputRequestQueueLock);
    mOutputRequestQueue.removeItemsAt(0);
}

void CameraHardware3v2::queueFrameNumber(int32_t num)
{
    Mutex::Autolock l(mFrameNumberQueueLock);
    mFrameNumberQueue.push_back(num);
}

void CameraHardware3v2::dequeueFrameNumber()
{
    Mutex::Autolock l(mFrameNumberQueueLock);
    ALOGV("%s[%d]: mFrameNumberQueue: %d", __FUNCTION__, __LINE__, mFrameNumberQueue[0]);
    mFrameNumberQueue.removeItemsAt(0);
}

void CameraHardware3v2::queueReprocessRequest(const Request& r)
{
    {
        Mutex::Autolock l(mInputRequestQueueLock);
        Request* pr = new Request();
        *pr = r;
        mInputRequestQueue.push_back(pr);
    }
    // new request received, notify shutter thread
    mShutterThread->signal();
}

void CameraHardware3v2::dequeueReprocessRequest()
{
    Mutex::Autolock l(mInputRequestQueueLock);
    mInputRequestQueue.removeItemsAt(0);
}

void CameraHardware3v2::submitReprocessRequest(Request* r)
{
    LOG_FUNCTION_NAME();
    mResultFrameNumber = r->frameNumber;

    char path[64] = {'\0'};
    // first , dump informations we intersted.
    int frameNumber = r->frameNumber;
    int output_bufs = r->num_output_buffers;

    // get input buffer Info
    int input_w = r->input_buffer->stream->width;
    int input_h = r->input_buffer->stream->height;
    int input_format = r->input_buffer->stream->format;
#ifdef HAL3_HWC_BUFFER
    int input_size = getStreamBufferSize(r->input_buffer->buffer);
#else
    ALOGV("%s[%d]: input buffer stream:[%dx%d], fmt:%d, input:[%dx%d], fmt:%d", __FUNCTION__, __LINE__,
          r->input_buffer->stream->width, r->input_buffer->stream->height, r->input_buffer->stream->format, input_w, input_h, input_format);
    int input_size = getStreamBufferSize(r->input_buffer->buffer, r->input_buffer->stream->width, r->input_buffer->stream->height, r->input_buffer->stream->format);
#endif
    void* input_vaddr = NULL;
    void* cb = NULL;
    void* cr = NULL;
#ifdef HAL3_HWC_BUFFER
    getStreamBufferVAddr(r->input_buffer->buffer, &input_vaddr, input_format, &cb, &cr);
#else
    getStreamBufferVAddr(r->input_buffer->buffer, &input_vaddr, input_w, input_h, r->input_buffer->stream->usage, input_format, &cb, &cr);
#endif
    uint32_t input_phyAddr = 0;
    getStreamBufferPhyAddr(r->input_buffer->buffer, &input_phyAddr);

    // collect misc info
    MISC_FIELD_INFO miscInfo;
    memset(&miscInfo, 0x0, sizeof(MISC_FIELD_INFO));
    collectRequestMiscInfo(
        r->settings,
        &miscInfo);

    // Wrong way to release input_buffer fence
#if 0
    // release input buffer fence
    r->input_buffer->acquire_fence = -1;
    r->input_buffer->release_fence = -1;
#endif

#if 0
    if(r->input_buffer_sync_fd >= 0) {
        sw_sync_timeline_inc(r->input_buffer_sync_fd, 1);
        close(r->input_buffer_sync_fd);
    }
#endif

    // update timestamp
    camera_metadata_entry_t e;
    e = r->settings.find(ANDROID_SENSOR_TIMESTAMP);
    //ALOGV("[%s] pts field count %d",__FUNCTION__,e.count);
    nsecs_t old_pts = 0;
    if(e.count > 0) {
        old_pts = e.data.i64[0];
        //ALOGV("[%s] org pts %lld",__FUNCTION__,old_pts);
    }

    // process and release output buffer fence
    HalBufferVector::iterator buf = r->buffers->begin();
    while(buf != r->buffers->end()) {
        int format = buf->stream->format;
        buffer_handle_t* pHandle = buf->buffer;
        int w = buf->stream->width;
        int h = buf->stream->height;
#ifdef HAL3_HWC_BUFFER
        int size = getStreamBufferSize(pHandle);
#else
        int size = getStreamBufferSize(pHandle, buf->stream->width, buf->stream->height, buf->stream->format);
#endif
        void* vaddr = NULL;
        void* cb = NULL;
        void* cr = NULL;
#ifdef HAL3_HWC_BUFFER
        getStreamBufferVAddr(pHandle, &vaddr, format, &cb, &cr);
#else
        getStreamBufferVAddr(pHandle, &vaddr, w, h, buf->stream->usage, format, &cb, &cr);
#endif
        uint32_t phyAddr = 0;
        getStreamBufferPhyAddr(pHandle, &phyAddr);

        // process BLOB buffer
        if(format == ANDROID_SCALER_AVAILABLE_FORMATS_BLOB) {

#if HAL3_HWC_BUFFER
            JPEG_SRC_BUF_INFO src;
            JPEG_DST_BUF_INFO dst;

            memset(&src, 0x0, sizeof(JPEG_SRC_BUF_INFO));
            memset(&dst, 0x0, sizeof(JPEG_DST_BUF_INFO));

            // prepare src info
            src.w = input_w;
            src.h = input_h;
            src.size = input_size;
            src.vaddr = input_vaddr;
            src.phyAddr = input_phyAddr;

            // prepare dst info
            dst.size = size;
            dst.w = w;
            dst.h = h;
            dst.vaddr = vaddr;
            dst.phyAddr = phyAddr;
            dst.bytesused = 0;      // set by V4L2

            dst.tw = miscInfo.thumbWidth;
            dst.th = miscInfo.thumbHeight;
            dst.thumbVAddr = NULL;
            dst.thumbBufUsed = 0;

            ALOGV("%s BLOB dst vAddr %p",
                  __FUNCTION__,
                  dst.vaddr);

            mFormatConvertHelper->compressJPEG(&src, &dst);

            // then append exif
            JPEG_RST_INFO jpgInfo;
            jpgInfo.vaddr = (uint8_t*) dst.vaddr;
            jpgInfo.size = dst.size;
            jpgInfo.jpegSize = dst.bytesused;
            jpgInfo.frameNumber = frameNumber;
            jpgInfo.pts = old_pts;
            jpgInfo.mSuccess = true;
            jpgInfo.mHasJPEG = true;
            jpgInfo.pThumbBuf = (uint8_t*) dst.thumbVAddr;
            jpgInfo.thumbSize = dst.thumbBufUsed;

            if(jpgInfo.jpegSize) {
                int32_t exif_data_estimate_size = (64 * 1024);
                int32_t wrkBufSize = jpgInfo.jpegSize + exif_data_estimate_size;
                uint8_t* wrkBuf = (uint8_t*) malloc(wrkBufSize);
                ALOGD("%s wrkBuf %p [%d vs %d]",
                      __FUNCTION__,
                      wrkBuf,
                      wrkBufSize,
                      dst.size);
                CHECK(wrkBuf != NULL);
                addExif(&jpgInfo, wrkBuf, &miscInfo);
                free(wrkBuf);
            }
#else
            BUF_CONVERT_INFO src;
            BUF_CONVERT_INFO dst;

            memset(&src, 0x0, sizeof(BUF_CONVERT_INFO));
            memset(&dst, 0x0, sizeof(BUF_CONVERT_INFO));

            // prepare src info
            src.w = input_w;
            src.h = input_h;
            src.bufferSize = input_size;
            src.vAddr = input_vaddr;
            src.format = input_format;

            // prepare dst info
            dst.w = w;
            dst.h = h;
            dst.vAddr = vaddr;
            dst.bufferSize = size;
            dst.format = format;
            dst.pHandle = pHandle;
            ALOGV("before compressToJPEG,  src.vAddr %p size %ix%i,dst.vAddr %p,dst size %ix%i ,thumb size %ix%i"
                  , src.vAddr, src.w, src.h, dst.vAddr, dst.w, dst.h, miscInfo.thumbWidth, miscInfo.thumbHeight);
            mFormatConvertHelper->compressToJPEG(&src, &dst, &miscInfo);
#endif
        }

        int status = CAMERA3_BUFFER_STATUS_OK;

        buf->status = status;
        // release fence
        buf->acquire_fence = -1;
        buf->release_fence = -1;
        ++buf;
    }

    camera3_capture_result result;
    memset(&result, 0x0, sizeof(camera3_capture_result));

    // construct result from mCurrentRequest info
    result.frame_number = r->frameNumber;
    result.result = r->settings.getAndLock();
    result.num_output_buffers = r->buffers->size();
    result.output_buffers = r->buffers->array();
    result.input_buffer = r->input_buffer;
    result.partial_result = 1;

    mOps->process_capture_result(mOps, &result);

    // send a notify
    camera3_notify_msg_t msg;
    msg.type = CAMERA3_MSG_SHUTTER;
    msg.message.shutter.frame_number = r->frameNumber;
    msg.message.shutter.timestamp = old_pts;

    ALOGV("%s[%d]: notify %d", __FUNCTION__, __LINE__, r->frameNumber);

    mOps->notify(mOps, &msg);

    // clean up
    r->settings.unlock(result.result);

    if(r->buffers != NULL) {
        delete r->buffers;
        r->buffers = NULL;
    }

    if(r->sensorBuffers != NULL ) {    //if (!needJpeg) {
        delete r->sensorBuffers;
        r->sensorBuffers = NULL;
    }

    r->settings.clear();

    if(r->input_buffer != NULL) {
        ALOGV("free input_buffer");
        free(r->input_buffer);
        r->input_buffer = NULL;
    }
    ALOGV("mFrameNumberQueue %d", mFrameNumberQueue[0]);

    // don't forget to update frameNumber
    {
        Mutex::Autolock l(mFrameNumberQueueLock);
        int32_t firstFrameNumber = mFrameNumberQueue[0];
        CHECK(firstFrameNumber == (int32_t)r->frameNumber);
    }

    dequeueFrameNumber();
}

// ======================== ShutterThread ======================= //
CameraHardware3v2::ShutterThread::ShutterThread(CameraHardware3v2* parent)
    : mQueueIndex(NO_QUEUE), mParent(parent), mRunning(true)
{
    LOG_FUNCTION_NAME();
}

CameraHardware3v2::ShutterThread::~ShutterThread()
{
    LOG_FUNCTION_NAME();
}

void CameraHardware3v2::ShutterThread::signal()
{
    Mutex::Autolock l(mThreadLock);
    mSignal.signal();
}

void CameraHardware3v2::ShutterThread::stop()
{
    mRunning = false;
    signal();
}

bool CameraHardware3v2::ShutterThread::threadLoop()
{
    LOG_FUNCTION_NAME();
    while( mParent->mOutputRequestQueue.size() <= 0
           && mParent->mInputRequestQueue.size() <= 0
           && mRunning) {

        Mutex::Autolock l(mThreadLock);
        mSignal.waitRelative(mThreadLock, kWaitNs);
    }
    if(!mRunning) {
        return false;
    }

    int32_t nextResultNumber = -1;

    {
        // find next request, frameNumber should be sequencial.
        Mutex::Autolock l(mParent->mFrameNumberQueueLock);
        nextResultNumber = mParent->mFrameNumberQueue[0];
    }

    Request* rst = NULL;

    mQueueIndex = NO_QUEUE;

    {
        Mutex::Autolock l(mParent->mOutputRequestQueueLock);
        ALOGV("set OUTPUT_QUEUE");
        // find matched Request in mOutputRequestQueue
        if(mParent->mOutputRequestQueue.size() > 0) {
            ALOGV("set OUTPUT_QUEUE 1 mParent->mOutputRequestQueue[0]->frameNumber %d , nextResultNumber %d", mParent->mOutputRequestQueue[0]->frameNumber, nextResultNumber);
            if((int)mParent->mOutputRequestQueue[0]->frameNumber == nextResultNumber) {
                rst = mParent->mOutputRequestQueue[0];
                mQueueIndex = OUTPUT_QUEUE;
            }
        }
    }
    {
        Mutex::Autolock l(mParent->mInputRequestQueueLock);
        // find matched Request in mInputRequestQueue
        if(mParent->mInputRequestQueue.size() > 0) {
            if((int)mParent->mInputRequestQueue[0]->frameNumber == nextResultNumber) {
                rst = mParent->mInputRequestQueue[0];
                mQueueIndex = INPUT_QUEUE;
            }
        }
    }
    if (rst == NULL) {
        ALOGV("ShutterThread::threadLoop  rst:[%p], %s: rst is NULL", rst, (mQueueIndex == OUTPUT_QUEUE) ? "Output" : "Input");
    } else {
        ALOGV("ShutterThread::threadLoop  rst:[%p], %s: rst->frameNumber:%d", rst, (mQueueIndex == OUTPUT_QUEUE) ? "Output" : "Input", rst->frameNumber);
    }
    if (rst == NULL || mQueueIndex == NO_QUEUE) {
#if 0
        int firstInputR = -1;
        if(mParent->mInputRequestQueue.size() > 0) {
            firstInputR = mParent->mInputRequestQueue[0]->frameNumber;
        }

        ALOGE("ShutterThread : ERR nextResultNumber %d InputQueueSize %d OutputQueueSize %d first input request %d",
              nextResultNumber,
              mParent->mInputRequestQueue.size(),
              mParent->mOutputRequestQueue.size(),
              firstInputR);
#endif
        usleep(10);
        if(mParent->mResultFrameNumber >= mParent->mRequestFrameNumber) {
            ALOGV("signalDeviceIdle, mResultFrameNumber:%i, mRequestFrameNumber:%i", mParent->mResultFrameNumber, mParent->mRequestFrameNumber);
            mParent->signalDeviceIdle();
        }
        return true;
    }

    CHECK(rst != NULL);
    CHECK(mQueueIndex != NO_QUEUE);
#if (ADJUST_RESULT_FLOW)
    // for cts -> add flow control mechanism
    nsecs_t captureSystemTime = systemTime();
    nsecs_t captureTime = rst->pts;

    if(mParent->mPrevPTS > 0 && mParent->mPrevSystemTime > 0) {
        int64_t ptsDiff = captureTime - mParent->mPrevPTS;
        int64_t stDiff = (int64_t) (captureSystemTime - mParent->mPrevSystemTime);

        LOGVV("DBG7 ptsDiff %lld stDiff %lld", ptsDiff, stDiff);
        int64_t diff = ptsDiff - stDiff;
        if(diff > 1E9) {
            LOGVV("DBG7 Request delay %.2f sec, reset", (float)diff / 1E9f);
            // someone stop the request but does not stop camera.
            // special case
            diff = 0;
            mParent->mPrevPTS = 0;
            mParent->mPrevSystemTime = 0;
        }

        if(diff > 0) {
            LOGVV("DBG7 pts gap to large %.2f ms", (float)diff / 1E6f);
            usleep( (diff / 2) / 1E3);
        }
    }

    mParent->mPrevPTS = captureTime;
    mParent->mPrevSystemTime = captureSystemTime;
#endif
    ALOGV("ShutterThread::threadLoop  mQueueIndex:%d", mQueueIndex);
    switch(mQueueIndex) {
        case OUTPUT_QUEUE:
            if(rst->mSuccess) {
                mParent->submitRequest(rst);
            } else {
                mParent->submitErrorRequest(rst);
            }
            delete rst;
            // remove this request
            mParent->dequeueOutputRequest();
            break;
        case INPUT_QUEUE:
            mParent->submitReprocessRequest(rst);
            delete rst;
            mParent->dequeueReprocessRequest();
            break;
        case NO_QUEUE:
        default:
            break;
    }

    if(mParent->mResultFrameNumber >= mParent->mRequestFrameNumber) {
        ALOGD("signalDeviceIdle");
        mParent->signalDeviceIdle();
    }

    return true;
}

}; // namespace android
